////////////////////////////////////////////////////////////////////////////////
//
//  TYPHOON FRAMEWORK
//  Copyright 2013, Typhoon Framework Contributors
//  All Rights Reserved.
//
//  NOTICE: The authors permit you to use, modify, and distribute this file
//  in accordance with the terms of the license agreement accompanying it.
//
////////////////////////////////////////////////////////////////////////////////


#import <XCTest/XCTest.h>
#import "Typhoon.h"
#import "TyphoonStoryboard.h"
#import "StoryboardViewControllerAssembly.h"
#import "StoryboardFirstViewController.h"
#import "StoryboardTabBarViewController.h"
#import "StoryboardPageViewController.h"
#import "StoryboardMainViewController.h"
#import "StoryboardDetailViewController.h"
#import "StoryboardWithReferenceAssembly.h"

@interface StoryboardTests : XCTestCase

@property (strong, nonatomic) id factory;
@property (strong, nonatomic) id factoryWithReference;

@end

@implementation StoryboardTests
{
    UIStoryboard *storyboard;
    UIStoryboard *storyboardWithReference;
}

- (void)setUp
{
    [super setUp];

    self.factory = [TyphoonBlockComponentFactory factoryWithAssembly:[StoryboardViewControllerAssembly assembly]];
    self.factoryWithReference = [TyphoonBlockComponentFactory factoryWithAssembly:[StoryboardWithReferenceAssembly assembly]];
    
    storyboard = [self.factory storyboard];
    storyboardWithReference = [self.factoryWithReference storyboardWithReference];
}

- (void)tearDown
{
    self.factory = nil;
    self.factoryWithReference = nil;
    [super tearDown];
}

- (void)test_initial
{
    UIViewController *controller = [storyboard instantiateInitialViewController];
    XCTAssertEqualObjects(controller.title, @"Initial");
}

- (void)test_first
{
    UIViewController *controller = [storyboard instantiateViewControllerWithIdentifier:@"first"];
    XCTAssertEqualObjects(controller.title, @"First");
}

- (void)test_first_resolves_circular_dependencies {
    StoryboardFirstViewController *controller = [storyboard instantiateViewControllerWithIdentifier:@"first"];
    StoryboardControllerDependency *dependency = controller.dependency;
    StoryboardFirstViewController *circular = dependency.circularDependencyBackToViewController;

    XCTAssertTrue(controller == circular);
}

- (void)test_first_from_navigation
{
    UINavigationController *controller = [storyboard instantiateViewControllerWithIdentifier:@"navigation"];
    XCTAssertEqualObjects([[controller.viewControllers firstObject] title], @"First");
}

- (void)test_second
{
    UIViewController *controller = [storyboard instantiateViewControllerWithIdentifier:@"second"];
    XCTAssertEqualObjects(controller.title, @"Second");
}

- (void)test_second_from_navigation
{
    UINavigationController *controller = [storyboard instantiateViewControllerWithIdentifier:@"navigation2"];
    XCTAssertEqualObjects([[controller.viewControllers firstObject] title], @"Second");
}

- (void)test_unique
{
    UINavigationController *controller = [storyboard instantiateViewControllerWithIdentifier:@"unique"];
    XCTAssertEqualObjects(controller.title, @"Unique");
}

- (void)test_singleton_controller
{
    UIViewController *factoryController = [self.factory singletonViewController];
    XCTAssertEqualObjects(factoryController.title, @"singleton");
    
    UIViewController *storyboardController = [storyboard instantiateViewControllerWithIdentifier:@"SingletonViewController"];
    XCTAssertEqualObjects(storyboardController.title, @"singleton");
    
    XCTAssertEqualObjects(factoryController, storyboardController);
}

- (void)test_lazy_singleton_controller
{
    UIViewController *storyboardController = [storyboard instantiateViewControllerWithIdentifier:@"LazySingletonViewController"];
    XCTAssertEqualObjects(storyboardController.title, @"lazysingleton");
    
    UIViewController *factoryController = [self.factory lazySingletonViewController];
    XCTAssertEqualObjects(factoryController.title, @"lazysingleton");
    
    XCTAssertEqualObjects(factoryController, storyboardController);
}

- (void)test_weak_singleton_controller
{
    UIViewController *storyboardController;
    UIViewController *factoryController;
    
    @autoreleasepool {
        storyboardController = [storyboard instantiateViewControllerWithIdentifier:@"WeakSingletonViewController"];
        XCTAssertEqualObjects(storyboardController.title, @"weaksingleton");
        
        storyboardController.title = @"step1";
        
        factoryController = [self.factory weakSingletonViewController];
        XCTAssertEqualObjects(factoryController.title, @"step1");
    }
    
    storyboardController = nil;
    factoryController = nil;
    
    UIViewController *otherController = [self.factory weakSingletonViewController];
    XCTAssertEqualObjects(otherController.title, @"weaksingleton");
}

- (void)test_prototype_controller
{
    UIViewController *firstInstance = [storyboard instantiateViewControllerWithIdentifier:@"PrototypeViewController"];
    UIViewController *secondInstance = [storyboard instantiateViewControllerWithIdentifier:@"PrototypeViewController"];
    XCTAssertNotEqualObjects(firstInstance, secondInstance);
}

- (void)test_controller_created_by_storyboard_name
{
    UIViewController *oneMoreViewController = [storyboard instantiateViewControllerWithIdentifier:@"OneMoreViewController"];
    XCTAssertEqualObjects(oneMoreViewController.title, @"OneMoreViewController");
}

- (void)test_controller_with_runtime_attributes {
    NSString *controllerId = @"OneMoreViewController";
    NSString *testTitle = @"Test Title";
    UIViewController *result = [self.factory oneMoreViewControllerWithId:controllerId title:testTitle];
    XCTAssertEqualObjects(result.title, testTitle);
}

- (void)test_count_of_models_in_controller {
    StoryboardMainViewController *mainController = [storyboardWithReference instantiateViewControllerWithIdentifier:@"StoryboardMainViewController"];
    [[UIApplication sharedApplication] keyWindow].rootViewController = mainController;
    [mainController performSegueWithIdentifier:@"detail" sender:nil];
    XCTAssertNotNil(mainController.presentedViewController);
    StoryboardDetailViewController *detailViewController = (StoryboardDetailViewController *) mainController.presentedViewController;
    XCTAssertEqual(detailViewController.countOfModelInstanceInjections, 1);
}

- (void)test_count_of_models_in_tab_controller {
    StoryboardTabBarViewController *tabController = [storyboardWithReference instantiateViewControllerWithIdentifier:@"StoryboardTabBarViewController"];
    
    NSLog(@"test_count_of_models_in_tab_controller");
    
    for (StoryboardPageViewController *viewController in tabController.viewControllers) {
        NSLog(@"viewController %@",viewController);
        NSLog(@"%ld",(long)viewController.countOfModelInstanceInjections);
        XCTAssertEqual(viewController.countOfModelInstanceInjections, 1);
    }
}


@end
